# -*- coding: utf-8 -*-
# 
# Copyright (c) 2025~2999 - Cologler <skyoflw@gmail.com>
# ----------
#
# ----------

SERVICEPROVIDER_NAMING_CONVENTION: tuple[str, ...] = (
    'ioc',
    'provider',
    'service_provider'
)
